# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc
import warnings

from openfl.protocols import aggregator_pb2 as openfl_dot_protocols_dot_aggregator__pb2
from openfl.protocols import base_pb2 as openfl_dot_protocols_dot_base__pb2

GRPC_GENERATED_VERSION = '1.65.5'
GRPC_VERSION = grpc.__version__
EXPECTED_ERROR_RELEASE = '1.66.0'
SCHEDULED_RELEASE_DATE = 'August 6, 2024'
_version_not_supported = False

try:
    from grpc._utilities import first_version_is_lower
    _version_not_supported = first_version_is_lower(GRPC_VERSION, GRPC_GENERATED_VERSION)
except ImportError:
    _version_not_supported = True

if _version_not_supported:
    warnings.warn(
        f'The grpc package installed is at version {GRPC_VERSION},'
        + f' but the generated code in openfl/protocols/aggregator_pb2_grpc.py depends on'
        + f' grpcio>={GRPC_GENERATED_VERSION}.'
        + f' Please upgrade your grpc module to grpcio>={GRPC_GENERATED_VERSION}'
        + f' or downgrade your generated code using grpcio-tools<={GRPC_VERSION}.'
        + f' This warning will become an error in {EXPECTED_ERROR_RELEASE},'
        + f' scheduled for release on {SCHEDULED_RELEASE_DATE}.',
        RuntimeWarning
    )


class AggregatorStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.Ping = channel.unary_unary(
                '/openfl.aggregator.Aggregator/Ping',
                request_serializer=openfl_dot_protocols_dot_aggregator__pb2.PingRequest.SerializeToString,
                response_deserializer=openfl_dot_protocols_dot_aggregator__pb2.PingResponse.FromString,
                _registered_method=True)
        self.GetTasks = channel.unary_unary(
                '/openfl.aggregator.Aggregator/GetTasks',
                request_serializer=openfl_dot_protocols_dot_aggregator__pb2.GetTasksRequest.SerializeToString,
                response_deserializer=openfl_dot_protocols_dot_aggregator__pb2.GetTasksResponse.FromString,
                _registered_method=True)
        self.GetAggregatedTensor = channel.unary_unary(
                '/openfl.aggregator.Aggregator/GetAggregatedTensor',
                request_serializer=openfl_dot_protocols_dot_aggregator__pb2.GetAggregatedTensorRequest.SerializeToString,
                response_deserializer=openfl_dot_protocols_dot_aggregator__pb2.GetAggregatedTensorResponse.FromString,
                _registered_method=True)
        self.SendLocalTaskResults = channel.stream_unary(
                '/openfl.aggregator.Aggregator/SendLocalTaskResults',
                request_serializer=openfl_dot_protocols_dot_base__pb2.DataStream.SerializeToString,
                response_deserializer=openfl_dot_protocols_dot_aggregator__pb2.SendLocalTaskResultsResponse.FromString,
                _registered_method=True)
        self.InteropRelay = channel.unary_unary(
                '/openfl.aggregator.Aggregator/InteropRelay',
                request_serializer=openfl_dot_protocols_dot_aggregator__pb2.InteropMessage.SerializeToString,
                response_deserializer=openfl_dot_protocols_dot_aggregator__pb2.InteropMessage.FromString,
                _registered_method=True)


class AggregatorServicer(object):
    """Missing associated documentation comment in .proto file."""

    def Ping(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetTasks(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetAggregatedTensor(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def SendLocalTaskResults(self, request_iterator, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def InteropRelay(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_AggregatorServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'Ping': grpc.unary_unary_rpc_method_handler(
                    servicer.Ping,
                    request_deserializer=openfl_dot_protocols_dot_aggregator__pb2.PingRequest.FromString,
                    response_serializer=openfl_dot_protocols_dot_aggregator__pb2.PingResponse.SerializeToString,
            ),
            'GetTasks': grpc.unary_unary_rpc_method_handler(
                    servicer.GetTasks,
                    request_deserializer=openfl_dot_protocols_dot_aggregator__pb2.GetTasksRequest.FromString,
                    response_serializer=openfl_dot_protocols_dot_aggregator__pb2.GetTasksResponse.SerializeToString,
            ),
            'GetAggregatedTensor': grpc.unary_unary_rpc_method_handler(
                    servicer.GetAggregatedTensor,
                    request_deserializer=openfl_dot_protocols_dot_aggregator__pb2.GetAggregatedTensorRequest.FromString,
                    response_serializer=openfl_dot_protocols_dot_aggregator__pb2.GetAggregatedTensorResponse.SerializeToString,
            ),
            'SendLocalTaskResults': grpc.stream_unary_rpc_method_handler(
                    servicer.SendLocalTaskResults,
                    request_deserializer=openfl_dot_protocols_dot_base__pb2.DataStream.FromString,
                    response_serializer=openfl_dot_protocols_dot_aggregator__pb2.SendLocalTaskResultsResponse.SerializeToString,
            ),
            'InteropRelay': grpc.unary_unary_rpc_method_handler(
                    servicer.InteropRelay,
                    request_deserializer=openfl_dot_protocols_dot_aggregator__pb2.InteropMessage.FromString,
                    response_serializer=openfl_dot_protocols_dot_aggregator__pb2.InteropMessage.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'openfl.aggregator.Aggregator', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))
    server.add_registered_method_handlers('openfl.aggregator.Aggregator', rpc_method_handlers)


 # This class is part of an EXPERIMENTAL API.
class Aggregator(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def Ping(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/openfl.aggregator.Aggregator/Ping',
            openfl_dot_protocols_dot_aggregator__pb2.PingRequest.SerializeToString,
            openfl_dot_protocols_dot_aggregator__pb2.PingResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetTasks(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/openfl.aggregator.Aggregator/GetTasks',
            openfl_dot_protocols_dot_aggregator__pb2.GetTasksRequest.SerializeToString,
            openfl_dot_protocols_dot_aggregator__pb2.GetTasksResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetAggregatedTensor(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/openfl.aggregator.Aggregator/GetAggregatedTensor',
            openfl_dot_protocols_dot_aggregator__pb2.GetAggregatedTensorRequest.SerializeToString,
            openfl_dot_protocols_dot_aggregator__pb2.GetAggregatedTensorResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def SendLocalTaskResults(request_iterator,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.stream_unary(
            request_iterator,
            target,
            '/openfl.aggregator.Aggregator/SendLocalTaskResults',
            openfl_dot_protocols_dot_base__pb2.DataStream.SerializeToString,
            openfl_dot_protocols_dot_aggregator__pb2.SendLocalTaskResultsResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def InteropRelay(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/openfl.aggregator.Aggregator/InteropRelay',
            openfl_dot_protocols_dot_aggregator__pb2.InteropMessage.SerializeToString,
            openfl_dot_protocols_dot_aggregator__pb2.InteropMessage.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)
